/* -*- c++ -*- */
/////////////////////////////////////////////////////////////////////////////
//
// Main.cpp  -- Copyright (c) 2006 David Henry
// last modification: feb. 3, 2006
//
// This code is licenced under the MIT license.
//
// This software is provided "as is" without express or implied
// warranties. You may freely copy and compile this source into
// applications you distribute provided that the copyright text
// below is included in the resulting source code.
//
// Texture module test application.
//
/////////////////////////////////////////////////////////////////////////////

#include <GL/glew.h>
#include <GL/glut.h>
#include <iostream>
#include <cstdlib>
#include <vector>
#include <dirent.h>
#include <sys/types.h>

#include "GlErrors.h"
#include "Texture.h"

using std::cout;
using std::cerr;
using std::endl;
using std::vector;

/* mouse */
struct mouse_input_t
{
  int buttons[3];
  int x, y;

} mouse;

/* timer */
struct glut_timer_t
{
  double curent_time;
  double last_time;

} timer;

/* camera */
struct Vector3d {
  float _x, _y, _z;

} rot, eye;

static int modifiers;

int texIndex = 0;
vector<Texture *> texList;


static void
loadTextures (const string &dirname)
{
  dirent *dit;

  // Open the directory
  DIR *dd = opendir (dirname.c_str ());
  if (!dd)
    throw std::runtime_error ("Couldn't open dir");

  cout << "------ reading dir ------" << endl;

  // Read directory for textures
  while ((dit = readdir (dd)) != NULL)
    {
      const string filename (dit->d_name);
      const string path (dirname + "/" + filename);

      const char *str = filename.c_str ();
      string::size_type l = filename.find_last_of ('.');

      // Skip directories
      if (l > filename.length ())
	continue;

      // Skip files beginning with "<char>_" and files
      // ending with "_i.<char*>"
      if ((str[1] != '_') &&
	  !((str[l-1] == 'i') && (str[l-2] == '_')))
	{
	  // Check if it's a known image file format
	  if (filename.compare (l, 4, ".pcx") == 0 ||
	      filename.compare (l, 4, ".tga") == 0 ||
	      filename.compare (l, 4, ".png") == 0 ||
	      filename.compare (l, 4, ".dds") == 0 ||
	      filename.compare (l, 4, ".jpg") == 0 ||
	      filename.compare (l, 4, ".PCX") == 0 ||
	      filename.compare (l, 4, ".TGA") == 0 ||
	      filename.compare (l, 4, ".PNG") == 0 ||
	      filename.compare (l, 4, ".DDS") == 0 ||
	      filename.compare (l, 4, ".JPG") == 0)
	    {
	      cout << "loading \"" << path << "\"" << endl;

	      Texture *tex = new Texture2D (path, Texture::kCompress);

	      if (!tex->fail ())
		texList.push_back (tex);

	      checkOpenGLErrors (path.c_str ());
	    }
	}
    }

  // Close directory
  closedir (dd);

  cout << "---------- done ----------" << endl << endl;

  /*
#if 0
  tex = new TextureRectangle (filename);
#elif 1
  tex = new Texture2D (filename, Texture::kCompress);
#else
  string base, ext;

  // Extract base name and file extension
  base.assign (filename, 0, filename.find_last_of ('.'));
  ext.assign (filename, filename.find_last_of ('.'), 4);

  std::vector<string> files;
  files.push_back (base + "_posx" + ext);
  files.push_back (base + "_negx" + ext);
  files.push_back (base + "_posy" + ext);
  files.push_back (base + "_negy" + ext);
  files.push_back (base + "_posz" + ext);
  files.push_back (base + "_negz" + ext);

  tex = new TextureCubeMap (filename, files);
#endif
  */
}


static void
init (const string &dirname)
{
  // Initialize mouse
  mouse.buttons[GLUT_LEFT_BUTTON] = GLUT_UP;
  mouse.buttons[GLUT_MIDDLE_BUTTON] = GLUT_UP;
  mouse.buttons[GLUT_RIGHT_BUTTON] = GLUT_UP;
  mouse.x = 0;
  mouse.y = 0;

  // Initialize timer
  timer.curent_time = 0;
  timer.last_time = 0;

  // Initialize camera input
  rot._x = 0.0f;   eye._x =  0.0f;
  rot._y = 0.0f;   eye._y =  0.0f;
  rot._z = 0.0f;   eye._z =  7.0f;

  // Initialize OpenGL
  glClearColor (0.5f, 0.5f, 0.5f, 1.0f);
  glShadeModel (GL_SMOOTH);

  glEnable (GL_DEPTH_TEST);
  glEnable (GL_TEXTURE_2D);
  glEnable (GL_TEXTURE_RECTANGLE_ARB);
  glEnable (GL_LIGHTING);
  glEnable (GL_LIGHT0);
  glEnable (GL_BLEND);

  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  // Initialize GLEW
  GLenum err = glewInit ();
  if (GLEW_OK != err)
    {
      // Problem: glewInit failed, something is seriously wrong.
      cerr << "Error: " << glewGetErrorString (err) << endl;
    }

  // Initialize texture
  //glPixelStorei (GL_PACK_SKIP_PIXELS, 150);
  //glPixelStorei (GL_UNPACK_SKIP_PIXELS, 64);
  //glPixelStorei (GL_UNPACK_ROW_LENGTH, 128);

  //glPixelTransferf (GL_RED_SCALE, 0.2);
  //glPixelTransferf (GL_GREEN_SCALE, 0.8);
  //glPixelTransferf (GL_BLUE_SCALE, 1.2);

  // Load textures from the given directory
  loadTextures (dirname);

  if (texList.size () == 0)
    {
      cout << "No texture found in " << dirname << endl;
      exit (-1);
    }

  checkOpenGLErrors ("init");
}


static void
shutdownApp (void)
{
  for (unsigned int i = 0; i < texList.size (); ++i)
    delete texList[i];
}


static void
reshape (int w, int h)
{
  if (h == 0)
    h = 1;

  glViewport (0, 0, (GLsizei)w, (GLsizei)h);

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (45.0, (GLfloat)w/(GLfloat)h, 0.1f, 1000.0f);

  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  glutPostRedisplay ();
}


static void
updateTimer (struct glut_timer_t *t)
{
  t->last_time = t->curent_time;
  t->curent_time = static_cast<double> (glutGet (GLUT_ELAPSED_TIME)) / 1000.0f;
}


void
drawEnvironment (void)
{
  static GLfloat xPlane[] = { 1.0f, 0.0f, 0.0f, 0.0f };
  static GLfloat yPlane[] = { 0.0f, 1.0f, 0.0f, 0.0f };
  static GLfloat zPlane[] = { 0.0f, 0.0f, 1.0f, 0.0f };

  glEnable (GL_TEXTURE_GEN_S);
  glEnable (GL_TEXTURE_GEN_T);
  glEnable (GL_TEXTURE_GEN_R);

  // Enable cube map texture
  glEnable (GL_TEXTURE_CUBE_MAP_ARB);
  texList[texIndex]->bind (GL_TEXTURE0);

  // Setup texture coord generation
  glTexGeni (GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGeni (GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
  glTexGeni (GL_R, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);

  glTexGenfv (GL_S, GL_OBJECT_PLANE, xPlane);
  glTexGenfv (GL_T, GL_OBJECT_PLANE, yPlane);
  glTexGenfv (GL_R, GL_OBJECT_PLANE, zPlane);

  glDisable (GL_DEPTH_TEST);

  // Draw cube map sphere
  glutSolidSphere (15.0f, 60, 30);

  glEnable (GL_DEPTH_TEST);

  glDisable (GL_TEXTURE_CUBE_MAP_ARB);

  glDisable (GL_TEXTURE_GEN_S);
  glDisable (GL_TEXTURE_GEN_T);
  glDisable (GL_TEXTURE_GEN_R);
}


static void
display (void)
{
  updateTimer (&timer);

  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glActiveTexture (GL_TEXTURE0);
  glMatrixMode (GL_TEXTURE);
  glLoadIdentity ();
  glMatrixMode (GL_MODELVIEW);
  glLoadIdentity ();

  //GLfloat lightPos[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
  //glLightfv (GL_LIGHT0, GL_POSITION, lightPos);

  //GLfloat lightColor[4] = { 0.7f, 0.3f, 0.3f, 1.0f };
  //glLightfv (GL_LIGHT0, GL_DIFFUSE, lightColor);

  // Camera rotation
  glTranslated (-eye._x, -eye._y, -eye._z);
  glRotated (rot._x, 1.0f, 0.0f, 0.0f);
  glRotated (rot._y, 0.0f, 1.0f, 0.0f);
  glRotated (rot._z, 0.0f, 0.0f, 1.0f);

  Texture *tex = texList[texIndex];

  GLint width, height;
  if (tex->target () == GL_TEXTURE_RECTANGLE_ARB)
    {
      width = dynamic_cast<TextureRectangle *>(tex)->width ();
      height = dynamic_cast<TextureRectangle *>(tex)->height ();
      glEnable (GL_TEXTURE_RECTANGLE_ARB);
      glDisable (GL_TEXTURE_2D);
    }
  else
    {
      width = height = 1;
      glEnable (GL_TEXTURE_2D);
      glDisable (GL_TEXTURE_RECTANGLE_ARB);
    }

  if (tex->target () != GL_TEXTURE_CUBE_MAP_ARB)
    {
      tex->bind (GL_TEXTURE0);

      if (!tex->stdCoordSystem ())
	{
	  //glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	  glActiveTexture (GL_TEXTURE0);
	  glMatrixMode (GL_TEXTURE);
	  glLoadIdentity ();
	  glScalef (1.0f, -1.0f, 1.0f);
	  glTranslatef (0.0f, -1.0f, 0.0f);
	  glMatrixMode (GL_MODELVIEW);
	}

#if 0
      glutSolidTeapot (1.0f);
#else
      // Draw quad
      glBegin (GL_QUADS);
	glTexCoord2f (0.0f, 0.0f);
	glVertex3f (-1.0f, -1.0f, 0.0f);

	glTexCoord2f (width, 0.0f);
	glVertex3f (1.0f, -1.0f, 0.0f);

	glTexCoord2f (width, height);
	glVertex3f (1.0f, 1.0f, 0.0f);

	glTexCoord2f (0.0f, height);
	glVertex3f (-1.0f, 1.0f, 0.0f);
      glEnd  ();
#endif
    }
  else
    {
      glDisable (GL_LIGHTING);
      drawEnvironment ();
    }

  glutSwapBuffers ();
}


static void
keyPress (unsigned char key, int x, int y)
{
  // Escape
  if (key == 27)
    exit (0);

  if (key == '+')
    texIndex++;

  if (key == '-')
    texIndex--;

  if (texIndex < 0)
    texIndex = 0;

  if (static_cast<unsigned int>(texIndex) == texList.size ())
    texIndex--;

  cout << "current texture: " << texList[texIndex]->name () << endl;

#if 0
  if (key == 's' || key == 'S')
    enable_shader = !enable_shader;
#endif

  glutPostRedisplay ();
}


static void
mouseMotion (int x, int y)
{
  if (mouse.buttons[GLUT_MIDDLE_BUTTON] == GLUT_DOWN)
    {
      // Zoom
      eye._z += (x - mouse.x) * 0.1;
    }
  else if (mouse.buttons[GLUT_LEFT_BUTTON] == GLUT_DOWN)
    {
      if (modifiers & GLUT_ACTIVE_SHIFT)
	{
	  // Translation
	  eye._x -= (x - mouse.x) * 0.02;
	  eye._y += (y - mouse.y) * 0.02;
	}
      else
	{
	  // Rotation
	  rot._x += (y - mouse.y);
	  rot._y += (x - mouse.x);
	}
    }

  mouse.x = x;
  mouse.y = y;

  glutPostRedisplay ();
}


static void
mouseButton (int button, int state, int x, int y)
{
  modifiers = glutGetModifiers();

  // Update mouse state
  mouse.buttons[button] = state;
  mouse.x = x;
  mouse.y = y;

  glutPostRedisplay ();
}


int
main (int argc, char *argv[])
{
  // Initialize glut
  glutInit (&argc, argv);

  if (argc < 2)
    {
      cerr << "usage: " << argv[0] << " <dirname>" << endl;
      return -1;
    }

  // Create glut window
  glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL);
  glutInitWindowSize (640, 480);
  glutCreateWindow ("Texture module");

  // Initialize application
  atexit (shutdownApp);
  init (argv[1]);

  // Setup glut callback functions
  glutReshapeFunc (reshape);
  glutDisplayFunc (display);
  glutKeyboardFunc (keyPress);
  glutMotionFunc (mouseMotion);
  glutMouseFunc (mouseButton);

  // Enter the main loop
  glutMainLoop();

  return 0;
}
